package com.jiuqi.grid;

import java.util.Stack;
import com.jiuqi.util.Base64;
import java.io.OutputStream;
import java.io.IOException;
import com.jiuqi.util.MemStream;
import java.io.InputStream;
import com.jiuqi.util.StreamException;
import com.jiuqi.util.Stream;
import java.util.List;
import com.jiuqi.util.StringUtils;
import java.util.Iterator;
import java.util.HashSet;
import java.util.Set;
import java.io.Serializable;

public final class GridData implements Serializable, Cloneable
{
    private IntList cellColors;
    private IntList fontColors;
    private IntList edgeColors;
    private GridFontNames fontNames;
    private IntList fontSizes;
    private IntList colWidths;
    private IntList rowHeights;
    private BitList colVisible;
    private BitList rowVisible;
    private BitList hideDataRow;
    private BitList hideDataCol;
    private BitList colAutoSize;
    private BitList rowAutoSize;
    private Matrix values;
    private Matrix cells;
    private GridFieldList rects;
    private GridCell cell;
    private PropList propList;
    private Matrix objects;
    private int scrollTopCol;
    private int scrollTopRow;
    private int scrollBottomCol;
    private int scrollBottomRow;
    private int curCol;
    private int curRow;
    private boolean useFontSize;
    private LabelItems labels;
    private Matrix cellCSS;
    private String gridDataClass;
    private String referenceBase;
    private Matrix cellScript;
    private int horzOffset;
    private int vertOffset;
    private boolean horzCenter;
    private boolean vertCenter;
    private ExtendedDatas extDatas;
    private int options;
    private Set modifiedPropIndexRecycle;
    private boolean modifying;
    private int colCountPerPage;
    private int rowCountPerPage;
    private boolean rowPagingFirst;
    public static final int GRID_AUTOMERGE = 1;
    public static final int GRID_AUTOSTYLE = 2;
    public static final int GRID_DYNAMICSTORE = 4;
    public static final int GRID_FULLSTYLE = 8;
    
    public GridData() {
        this.scrollTopCol = -1;
        this.scrollTopRow = -1;
        this.scrollBottomCol = -1;
        this.scrollBottomRow = -1;
        this.curCol = -1;
        this.curRow = -1;
        this.labels = new LabelItems();
        this.cellCSS = null;
        this.cellScript = null;
        this.modifying = false;
        this.colCountPerPage = 0;
        this.rowCountPerPage = 0;
        this.rowPagingFirst = true;
        this.cellColors = GridConsts.create64ColorsPallette();
        this.fontColors = GridConsts.create64ColorsPallette();
        this.edgeColors = GridConsts.create16ColorsPallette();
        (this.fontNames = new GridFontNames(64)).addFont("宋体");
        (this.fontSizes = new IntList()).add(9);
        this.colVisible = new BitList(true);
        this.rowVisible = new BitList(true);
        this.colAutoSize = new BitList(false);
        this.rowAutoSize = new BitList(false);
        this.hideDataRow = new BitList(false);
        this.hideDataCol = new BitList(false);
        this.colWidths = new IntList();
        this.rowHeights = new IntList();
        this.values = new Matrix();
        this.cells = new Matrix();
        this.objects = new Matrix();
        this.rects = new GridFieldList(this);
        this.cell = new GridCell();
        (this.propList = new PropList(20)).add(GridConsts.DEF_CELL_PROP);
        this.extDatas = new ExtendedDatas();
        this.setColCount(5);
        this.setRowCount(5);
        for (int i = 0; i < 5; ++i) {
            final GridCell cell = this.getCell(i, 0);
            this.setCell(i, 0, cell);
        }
        for (int i = 0; i < 5; ++i) {
            final GridCell cell = this.getCell(0, i);
            this.setCell(0, i, cell);
        }
        this.setOptions(this.options = 7);
    }
    
    public GridData(final int colCount, final int rowCount) {
        this();
        this.setColCount(colCount);
        this.setRowCount(rowCount);
    }
    
    public GridData(final int options) {
        this();
        this.setOptions(options);
    }
    
    public GridData(final int colCount, final int rowCount, final int options) {
        this(colCount, rowCount);
        this.setOptions(options);
    }
    
    public void beginModify() {
        if (this.isModifying()) {
            throw new IllegalStateException("无法对已经处于修改状态的GridData执行beginModify方法！");
        }
        if (this.modifiedPropIndexRecycle == null) {
            this.modifiedPropIndexRecycle = new HashSet();
        }
        this.modifying = true;
    }
    
    public void secureBeginModify() {
        synchronized (this) {
            this.beginModify();
        }
    }
    
    public boolean isModifying() {
        return this.modifying;
    }
    
    public void endModify() {
        this.rebuildPropIndexes(this.modifiedPropIndexRecycle);
        this.modifiedPropIndexRecycle.clear();
        this.modifying = false;
    }
    
    public void secureEndModify() {
        synchronized (this) {
            this.endModify();
        }
    }
    
    private void createCellCssMatrix() {
        if (this.cellCSS == null) {
            (this.cellCSS = new Matrix()).setSize(this.getColCount(), this.getRowCount());
        }
    }
    
    private void createCellScriptMatrix() {
        if (this.cellScript == null) {
            (this.cellScript = new Matrix()).setSize(this.getColCount(), this.getRowCount());
        }
    }
    
    private boolean checkCol(final int col, final boolean isInsert) {
        if (isInsert) {
            return col > 0 && col <= this.getColCount();
        }
        return col >= 0 && col < this.getColCount();
    }
    
    private boolean checkRow(final int row, final boolean isInsert) {
        if (isInsert) {
            return row > 0 && row <= this.getRowCount();
        }
        return row >= 0 && row < this.getRowCount();
    }
    
    private void rebuildPropIndex(final int col, final int row, final int oldIndex) {
        if (oldIndex < 0) {
            return;
        }
        if (this.modifying) {
            this.modifiedPropIndexRecycle.add(new Integer(oldIndex));
            return;
        }
        if (!this.cells.hasElement(new Integer(oldIndex))) {
            for (int x = 0; x < this.cells.getXCount(); ++x) {
                for (int y = 0; y < this.cells.getYCount(); ++y) {
                    Integer index = (Integer)this.cells.getElement(x, y);
                    if (index != null) {
                        if (index > oldIndex) {
                            index = new Integer(index - 1);
                            this.cells.setElement(x, y, index);
                        }
                    }
                }
            }
            this.propList.remove(oldIndex);
        }
    }
    
    private void rebuildPropIndexes(final Set bufferedIndex) {
        if (bufferedIndex.size() == 0 || this.propList.count() == 0) {
            return;
        }
        final int[] propCase = new int[this.propList.count()];
        for (final Object index : bufferedIndex) {
            if (!this.cells.hasElement(index)) {
                propCase[index] = -1;
            }
        }
        for (int ci = propCase.length - 1; ci >= 0; --ci) {
            if (propCase[ci] != 0) {
                this.propList.remove(ci);
            }
        }
        for (int ci = 1; ci < propCase.length; ++ci) {
            final int[] array = propCase;
            final int n = ci;
            array[n] += propCase[ci - 1];
        }
        if (propCase[propCase.length - 1] == 0) {
            return;
        }
        for (int x = 0; x < this.cells.getXCount(); ++x) {
            for (int y = 0; y < this.cells.getYCount(); ++y) {
                Integer index2 = (Integer)this.cells.getElement(x, y);
                if (index2 != null) {
                    if (propCase[index2] != 0) {
                        index2 = new Integer(index2 + propCase[index2]);
                        this.cells.setElement(x, y, index2);
                    }
                }
            }
        }
    }
    
    public String getDataClass() {
        return this.gridDataClass;
    }
    
    public void setDataClass(final String value) {
        this.gridDataClass = value;
    }
    
    public String getReferenceBase() {
        return this.referenceBase;
    }
    
    public void setReferenceBase(final String value) {
        this.referenceBase = value;
    }
    
    public String getCellData(final int col, final int row) {
        return (String)this.values.getElement(col, row);
    }
    
    public String getCellData(final String pos) {
        final CellPostion p = new CellPostion(pos);
        return this.getCellData(p.getCol(), p.getRow());
    }
    
    public void setCellData(final int col, final int row, final String value) {
        this.values.setElement(col, row, value);
    }
    
    public void setCellData(final String pos, final String value) {
        final CellPostion p = new CellPostion(pos);
        this.setCellData(p.getCol(), p.getRow(), value);
    }
    
    public String getCellClass(final int col, final int row) {
        return (this.cellCSS == null) ? null : ((String)this.cellCSS.getElement(col, row));
    }
    
    public void setCellClass(final int col, final int row, final String value) {
        if (this.cellCSS == null) {
            if (value == null || value.length() == 0) {
                return;
            }
            this.createCellCssMatrix();
        }
        this.cellCSS.setElement(col, row, value);
    }
    
    public boolean getUseFontSize() {
        return this.useFontSize;
    }
    
    public void setUserFontSize(final boolean value) {
        this.useFontSize = value;
    }
    
    public String getCellScript(final int col, final int row) {
        return (this.cellScript == null) ? null : ((String)this.cellScript.getElement(col, row));
    }
    
    public void setCellScript(final int col, final int row, final String value) {
        if (this.cellScript == null) {
            if (value == null || value.length() == 0) {
                return;
            }
            this.createCellScriptMatrix();
        }
        this.cellScript.setElement(col, row, value);
    }
    
    public int getCellStyleIndex(final int col, final int row) {
        final Integer index = (Integer)this.cells.getElement(col, row);
        return (index == null) ? 0 : index;
    }
    
    public GridCell[] getCellSytles() {
        if (this.propList.count() > 0) {
            final GridCell[] ret = new GridCell[this.propList.count()];
            for (int i = 0; i < ret.length; ++i) {
                (ret[i] = new GridCell()).init(this, 0, 0, this.propList.get(i));
            }
            return ret;
        }
        return new GridCell[0];
    }
    
    public GridCell[] getCellStyles(final int left, final int top, final int right, final int bottom, final boolean fullMode) {
        if (this.propList.count() == 0) {
            return new GridCell[0];
        }
        final GridCell[] ret = new GridCell[this.propList.count()];
        for (int col = left; col <= right; ++col) {
            for (int row = top; row <= bottom; ++row) {
                if (StringUtils.isEmpty(this.getCellClass(col, row))) {
                    final int index = this.getCellStyleIndex(col, row);
                    if (ret[index] == null) {
                        ret[index] = new GridCell();
                        if (fullMode) {
                            ret[index].init(this, col, row);
                        }
                        else {
                            ret[index].init(this, 0, 0, this.getCellPropData(col, row));
                        }
                    }
                }
            }
        }
        return ret;
    }
    
    public byte[] getCellPropData(final int col, final int row) {
        final Integer index = (Integer)this.cells.getElement(col, row);
        if (index != null) {
            return this.propList.get(index);
        }
        if (col == 0 || row == 0) {
            return GridConsts.DEF_HEADCELL_PROP;
        }
        return GridConsts.DEF_CELL_PROP;
    }
    
    public GridCell getCell(final int col, final int row) {
        this.cell.init(this, col, row);
        return this.cell;
    }
    
    public GridCell getCell(final String pos) {
        final CellPostion p = new CellPostion(pos);
        return this.getCell(p.getCol(), p.getRow());
    }
    
    public GridCell getCellEx(final int col, final int row) {
        final GridCell result = new GridCell();
        result.init(this, col, row);
        return result;
    }
    
    public GridCell getCellEx(final String pos) {
        final CellPostion p = new CellPostion(pos);
        return this.getCellEx(p.getCol(), p.getRow());
    }
    
    public GridCellProperty getCellForChange(final int col, final int row) {
        if (!this.checkCol(col, false) || !this.checkRow(row, false)) {
            return null;
        }
        final GridCellProperty result = new GridCellProperty();
        result.init(this, col, row);
        return result;
    }
    
    public GridCellProperty getCellForChange(final String pos) {
        final CellPostion p = new CellPostion(pos);
        return this.getCellForChange(p.getCol(), p.getRow());
    }
    
    public void setCell(final int col, final int row, final GridCell value) {
        int index = this.propList.indexOf(value.getPropData());
        if (index < 0) {
            index = this.propList.add(value.getPropData());
        }
        final boolean dynamicStore = (this.options & 0x4) != 0x0;
        final Integer oldIndex = dynamicStore ? ((Integer)this.cells.getElement(col, row)) : null;
        this.cells.setElement(col, row, new Integer(index));
        if (dynamicStore && oldIndex != null && oldIndex != index) {
            this.rebuildPropIndex(col, row, oldIndex);
        }
    }
    
    public void setCell(final GridCell value) {
        this.setCell(value.getColNum(), value.getRowNum(), value);
    }
    
    public void setCell(final int left, final int top, final int right, final int bottom, final GridCell value) {
        this.setCell(new CellField(left, top, right, bottom), value);
    }
    
    public void setCell(final CellField field, final GridCell value) {
        int index = this.propList.indexOf(value.getPropData());
        if (index == -1) {
            index = this.propList.add(value.getPropData());
        }
        final Integer p = new Integer(index);
        for (int col = field.left; col <= field.right; ++col) {
            for (int row = field.top; row <= field.bottom; ++row) {
                final CellField f = this.rects.getMergeRect(col, row);
                if (f != null) {
                    if (f.left != col) {
                        continue;
                    }
                    if (f.top != row) {
                        continue;
                    }
                }
                this.cells.setElement(col, row, p);
            }
        }
    }
    
    public void setCell(final CellField[] fields, final GridCell value) {
        if (fields == null || fields.length == 0) {
            return;
        }
        int index = this.propList.indexOf(value.getPropData());
        if (index == -1) {
            index = this.propList.add(value.getPropData());
        }
        final Integer p = new Integer(index);
        for (int i = 0; i < fields.length; ++i) {
            for (int col = fields[i].left; col <= fields[i].right; ++col) {
                for (int row = fields[i].top; row <= fields[i].bottom; ++row) {
                    final CellField f = this.rects.getMergeRect(col, row);
                    if (f != null) {
                        if (f.left != col) {
                            continue;
                        }
                        if (f.top != row) {
                            continue;
                        }
                    }
                    this.cells.setElement(col, row, p);
                }
            }
        }
    }
    
    public Object getObj(final int col, final int row) {
        return this.objects.getElement(col, row);
    }
    
    public void setObj(final int col, final int row, final Object obj) {
        this.objects.setElement(col, row, obj);
    }
    
    public boolean getColAutoSize(final int col) {
        return this.colAutoSize.get(col);
    }
    
    public void setColAutoSize(final int col, final boolean value) {
        this.colAutoSize.set(col, value);
    }
    
    public boolean getColVisible(final int col) {
        return this.colVisible.count() <= 0 || this.colVisible.get(col);
    }
    
    public void setColVisible(final int col, final boolean value) {
        if (!value || col < this.colVisible.count()) {
            this.colVisible.set(col, value);
        }
    }
    
    public int getColCount() {
        return this.colWidths.count();
    }
    
    public int getColWidths(final int col) {
        return this.colWidths.get(col);
    }
    
    public void setColWidths(final int col, final int value) {
        this.colWidths.set(col, value);
    }
    
    public boolean getRowAutoSize(final int row) {
        return this.rowAutoSize.get(row);
    }
    
    public void setRowAutoSize(final int row, final boolean value) {
        this.rowAutoSize.set(row, value);
    }
    
    public boolean getRowVisible(final int row) {
        return this.rowVisible.count() <= 0 || this.rowVisible.get(row);
    }
    
    public void setRowVisible(final int row, final boolean value) {
        if (!value || row < this.getRowCount()) {
            this.rowVisible.set(row, value);
        }
    }
    
    @Deprecated
    public boolean getHideDataRow(final int row) {
        return this.hideDataRow.get(row);
    }
    
    @Deprecated
    public void setHideDataRow(final int row, final boolean value) {
        if (!value || row < this.getRowCount()) {
            this.hideDataRow.set(row, value);
        }
    }
    
    @Deprecated
    public boolean getHideDataCol(final int col) {
        return this.hideDataCol.get(col);
    }
    
    @Deprecated
    public int[][] getHideDataCols() {
        int[][] result = null;
        int count = 0;
        boolean state = false;
        for (int i = 1; i < this.getColCount(); ++i) {
            if (!state && this.getHideDataCol(i)) {
                ++count;
                state = true;
            }
            if (state && !this.getHideDataCol(i)) {
                state = false;
            }
        }
        if (count > 0) {
            result = new int[2][count];
            state = false;
            int k = 0;
            for (int j = 1; j < this.getColCount(); ++j) {
                if (!state && this.getHideDataCol(j)) {
                    result[0][k] = j;
                    state = true;
                }
                if (state && !this.getHideDataCol(j)) {
                    state = false;
                    result[1][k] = j - 1;
                    ++k;
                }
            }
            if (state) {
                result[1][count - 1] = this.getColCount() - 1;
            }
        }
        return result;
    }
    
    @Deprecated
    public int[][] getHideDataRows() {
        int[][] result = null;
        int count = 0;
        boolean state = false;
        for (int i = 1; i < this.getRowCount(); ++i) {
            if (!state && this.getHideDataRow(i)) {
                ++count;
                state = true;
            }
            if (state && !this.getHideDataRow(i)) {
                state = false;
            }
        }
        if (count > 0) {
            result = new int[2][count];
            state = false;
            int k = 0;
            for (int j = 1; j < this.getRowCount(); ++j) {
                if (!state && this.getHideDataRow(j)) {
                    result[0][k] = j;
                    state = true;
                }
                if (state && !this.getHideDataRow(j)) {
                    state = false;
                    result[1][k] = j - 1;
                    ++k;
                }
            }
            if (state) {
                result[1][count - 1] = this.getRowCount() - 1;
            }
        }
        return result;
    }
    
    @Deprecated
    public void setHideDataCol(final int col, final boolean value) {
        if (!value || col < this.getColCount()) {
            this.hideDataCol.set(col, value);
        }
    }
    
    public int getRowCount() {
        return this.rowHeights.count();
    }
    
    public void setColCount(final int value) {
        int colCount = this.getColCount();
        if (value > 0 && value < colCount) {
            final int delCount = colCount - value;
            this.colVisible.delete(value, delCount);
            this.colAutoSize.delete(value, delCount);
            this.colWidths.delete(value, delCount);
            this.values.xDelete(value, delCount);
            this.cells.xDelete(value, delCount);
            this.objects.xDelete(value, delCount);
            if (this.cellCSS != null) {
                this.cellCSS.xDelete(value, delCount);
            }
            if (this.cellScript != null) {
                this.cellScript.xDelete(value, delCount);
            }
            this.rects.delCols(value, delCount);
        }
        else if (value > colCount) {
            final int insCount = value - colCount;
            this.colVisible.insert(colCount, insCount);
            this.colAutoSize.insert(colCount, insCount);
            this.colWidths.insert(colCount, insCount, (colCount == 0) ? 100 : this.colWidths.get(colCount - 1));
            if (colCount == 0) {
                this.colWidths.set(0, 50);
            }
            this.values.xInsert(colCount, insCount);
            this.cells.xInsert(colCount, insCount);
            this.objects.xInsert(colCount, insCount);
            if (this.cellCSS != null) {
                this.cellCSS.xInsert(colCount, insCount);
            }
            if (this.cellScript != null) {
                this.cellScript.xInsert(colCount, insCount);
            }
            this.rects.insertCols(colCount, insCount);
            if (this.getRowCount() > 0) {
                for (int i = colCount; i < colCount + insCount; ++i) {
                    final GridCell cell = this.getCell(i, 0);
                    this.setCell(i, 0, cell);
                }
            }
        }
        colCount = this.getColCount();
        if (this.scrollTopCol >= colCount) {
            this.setScrollTopCol(colCount - 1);
        }
        if (this.scrollBottomCol >= colCount) {
            this.setScrollBottomCol(colCount - 1);
        }
    }
    
    public void setRowCount(final int value) {
        final int rowCount = this.getRowCount();
        if (value < rowCount) {
            final int delCount = rowCount - value;
            this.rowVisible.delete(value, delCount);
            this.rowAutoSize.delete(value, delCount);
            this.rowHeights.delete(value, delCount);
            this.values.yDelete(value, delCount);
            this.cells.yDelete(value, delCount);
            this.objects.yDelete(value, delCount);
            if (this.cellCSS != null) {
                this.cellCSS.yDelete(value, delCount);
            }
            if (this.cellScript != null) {
                this.cellScript.yDelete(value, delCount);
            }
            this.rects.delRows(value, delCount);
        }
        else if (value > this.getRowCount()) {
            final int insCount = value - rowCount;
            this.rowVisible.insert(rowCount, insCount);
            this.rowAutoSize.insert(rowCount, insCount);
            this.rowHeights.insert(rowCount, insCount, (rowCount == 0) ? 20 : this.rowHeights.get(rowCount - 1));
            this.values.yInsert(rowCount, insCount);
            this.objects.yInsert(rowCount, insCount);
            this.cells.yInsert(rowCount, insCount);
            if (this.cellCSS != null) {
                this.cellCSS.yInsert(rowCount, insCount);
            }
            if (this.cellScript != null) {
                this.cellScript.yInsert(rowCount, insCount);
            }
            this.rects.insertRows(rowCount, insCount);
            if (this.getColCount() > 0) {
                for (int i = rowCount; i < rowCount + insCount; ++i) {
                    final GridCell cell = this.getCell(0, i);
                    this.setCell(0, i, cell);
                }
            }
        }
    }
    
    public void insertCol(final int index, final int count) {
        this.insertCol(index, count, (this.options & 0x2) == 0x0, (this.options & 0x8) != 0x0);
    }
    
    private void insertCol(final int index, final int count, final boolean fastMode, final boolean fullMode) {
        if (!this.checkCol(index, true)) {
            return;
        }
        this.colVisible.insert(index, count);
        if (index > 0) {
            this.colAutoSize.insert(index, count, this.colAutoSize.get(index - 1));
        }
        else {
            this.colAutoSize.insert(index, count);
        }
        this.colWidths.insert(index, count, (this.getColCount() == 0 || index == 1) ? 100 : this.colWidths.get(index - 1));
        this.values.xInsert(index, count);
        this.cells.xInsert(index, count);
        this.objects.xInsert(index, count);
        if (this.cellCSS != null) {
            this.cellCSS.xInsert(index, count);
        }
        if (this.cellScript != null) {
            this.cellScript.xInsert(index, count);
        }
        this.rects.insertCols(index, count);
        if (this.getRowCount() > 0 && !fastMode) {
            for (int row = 1; row < this.getRowCount(); ++row) {
                GridCell cell;
                if (index > 1) {
                    cell = this.getCell(index - 1, row);
                }
                else {
                    cell = this.cell;
                    cell.init(this, -1, -1, GridConsts.DEF_CELL_PROP);
                }
                for (int col = index; col < index + count; ++col) {
                    this.setCell(col, row, cell);
                    if (fullMode) {
                        this.setCellClass(col, row, cell.getCssClass());
                        this.setCellScript(col, row, cell.getScript());
                    }
                }
            }
        }
        if (this.scrollTopCol >= index) {
            this.scrollTopCol += count;
        }
        if (this.scrollBottomCol + 1 >= index) {
            this.scrollBottomCol += count;
        }
    }
    
    public void deleteCol(final int index, int count) {
        if (!this.checkCol(index, false)) {
            return;
        }
        if (index + count > this.getColCount()) {
            count = this.getColCount() - index;
        }
        this.colVisible.delete(index, count);
        this.colAutoSize.delete(index, count);
        this.colWidths.delete(index, count);
        this.values.xDelete(index, count);
        this.cells.xDelete(index, count);
        this.objects.xDelete(index, count);
        if (this.cellCSS != null) {
            this.cellCSS.xDelete(index, count);
        }
        if (this.cellScript != null) {
            this.cellScript.xDelete(index, count);
        }
        this.rects.delCols(index, count);
        if (this.scrollTopCol >= index) {
            this.scrollTopCol = Math.min(Math.max(index, this.scrollTopCol - count), this.getColCount() - 1);
        }
        if (this.scrollBottomCol >= index) {
            this.scrollBottomCol = Math.min(Math.max(index, this.scrollBottomCol - count), this.getColCount() - 1);
        }
    }
    
    public void insertRow(final int index, final int count) {
        this.insertRow(index, count, (this.options & 0x2) == 0x0, (this.options & 0x8) != 0x0);
    }
    
    private void insertRow(final int index, final int count, final boolean fastMode, final boolean fullMode) {
        if (!this.checkRow(index, true)) {
            return;
        }
        this.rowVisible.insert(index, count);
        if (index > 0) {
            this.rowAutoSize.insert(index, count, this.rowAutoSize.get(index - 1));
        }
        else {
            this.rowAutoSize.insert(index, count);
        }
        this.rowHeights.insert(index, count, (this.getRowCount() == 0 || index == 1) ? 20 : this.rowHeights.get(index - 1));
        this.values.yInsert(index, count);
        this.cells.yInsert(index, count);
        this.objects.yInsert(index, count);
        if (this.cellCSS != null) {
            this.cellCSS.yInsert(index, count);
        }
        if (this.cellScript != null) {
            this.cellScript.yInsert(index, count);
        }
        this.rects.insertRows(index, count);
        if (this.getColCount() > 0 && !fastMode) {
            for (int col = 1; col < this.getColCount(); ++col) {
                GridCell cell;
                if (index > 1) {
                    cell = this.getCell(col, index - 1);
                }
                else {
                    cell = this.cell;
                    cell.init(this, -1, -1, GridConsts.DEF_CELL_PROP);
                }
                for (int row = index; row < index + count; ++row) {
                    this.setCell(col, row, cell);
                    if (fullMode) {
                        this.setCellClass(col, row, cell.getCssClass());
                        this.setCellScript(col, row, cell.getScript());
                    }
                }
            }
        }
        if (this.scrollTopRow >= index) {
            this.scrollTopRow += count;
        }
        if (this.scrollBottomRow + 1 >= index) {
            this.scrollBottomRow += count;
        }
    }
    
    public boolean duplicateCols(int fromIndex, int toIndex, final int duplicateCount, final int insertIndex) {
        if (!this.checkCol(fromIndex, false)) {
            return false;
        }
        if (!this.checkCol(toIndex, false)) {
            return false;
        }
        if (!this.checkCol(insertIndex, true)) {
            return false;
        }
        if (toIndex < fromIndex || duplicateCount <= 0 || (insertIndex > fromIndex && insertIndex <= toIndex)) {
            return false;
        }
        final int insertCount = (toIndex - fromIndex + 1) * duplicateCount;
        this.insertCol(insertIndex, insertCount, true, true);
        if (insertIndex <= fromIndex) {
            fromIndex += insertCount;
            toIndex += insertCount;
        }
        final List merges = (fromIndex < toIndex) ? this.rects.findFieldsInCols(fromIndex, toIndex) : null;
        for (int i = 0; i < duplicateCount; ++i) {
            final int offset = insertIndex + (toIndex - fromIndex + 1) * i - fromIndex;
            for (int j = fromIndex; j <= toIndex; ++j) {
                this.colWidths.set(offset + j, this.colWidths.get(j));
                this.values.xCopy(j, offset + j);
                if (this.cellCSS != null) {
                    this.cellCSS.xCopy(j, offset + j);
                }
                if (this.cellScript != null) {
                    this.cellScript.xCopy(j, offset + j);
                }
                this.cells.xCopy(j, offset + j);
                if (!this.colVisible.get(j)) {
                    this.colVisible.set(j + offset, false);
                }
            }
            if (merges != null && !merges.isEmpty()) {
                for (int j = 0; j < merges.size(); ++j) {
                    final CellField cf = merges.get(j);
                    this.mergeCells(cf.left + offset, cf.top, cf.right + offset, cf.bottom);
                }
            }
        }
        return true;
    }
    
    public boolean duplicateRows(int fromIndex, int toIndex, final int duplicateCount, final int insertIndex) {
        if (!this.checkRow(fromIndex, false)) {
            return false;
        }
        if (!this.checkRow(toIndex, false)) {
            return false;
        }
        if (!this.checkRow(insertIndex, true)) {
            return false;
        }
        if (toIndex < fromIndex || duplicateCount <= 0 || (insertIndex > fromIndex && insertIndex <= toIndex)) {
            return false;
        }
        final int insertCount = (toIndex - fromIndex + 1) * duplicateCount;
        this.insertRow(insertIndex, insertCount, true, true);
        if (insertIndex <= fromIndex) {
            fromIndex += insertCount;
            toIndex += insertCount;
        }
        final List merges = (fromIndex < toIndex) ? this.rects.findFieldsInRows(fromIndex, toIndex) : null;
        for (int i = 0; i < duplicateCount; ++i) {
            final int offset = insertIndex + (toIndex - fromIndex + 1) * i - fromIndex;
            for (int j = fromIndex; j <= toIndex; ++j) {
                this.rowHeights.set(offset + j, this.rowHeights.get(j));
                this.values.yCopy(j, offset + j);
                if (this.cellCSS != null) {
                    this.cellCSS.yCopy(j, offset + j);
                }
                if (this.cellScript != null) {
                    this.cellScript.yCopy(j, offset + j);
                }
                this.cells.yCopy(j, offset + j);
                if (!this.rowVisible.get(j)) {
                    this.rowVisible.set(j + offset, false);
                }
            }
            if (merges != null && !merges.isEmpty()) {
                for (int j = 0; j < merges.size(); ++j) {
                    final CellField cf = merges.get(j);
                    this.mergeCells(cf.left, cf.top + offset, cf.right, cf.bottom + offset);
                }
            }
        }
        return true;
    }
    
    public void deleteRow(final int index, int count) {
        if (!this.checkRow(index, false)) {
            return;
        }
        if (index + count > this.getRowCount()) {
            count = this.getRowCount() - index;
        }
        this.rowVisible.delete(index, count);
        this.rowAutoSize.delete(index, count);
        this.rowHeights.delete(index, count);
        this.values.yDelete(index, count);
        this.cells.yDelete(index, count);
        if (this.cellCSS != null) {
            this.cellCSS.yDelete(index, count);
        }
        if (this.cellScript != null) {
            this.cellScript.yDelete(index, count);
        }
        this.objects.yDelete(index, count);
        this.rects.delRows(index, count);
        if (this.scrollTopRow >= index) {
            this.scrollTopRow = Math.min(Math.max(index, this.scrollTopRow - count), this.getRowCount() - 1);
        }
        if (this.scrollBottomRow >= index) {
            this.scrollBottomRow = Math.min(Math.max(index, this.scrollBottomRow - count), this.getRowCount() - 1);
        }
    }
    
    public int getRowHeights(final int row) {
        return this.rowHeights.get(row);
    }
    
    public void setRowHeights(final int row, final int value) {
        this.rowHeights.set(row, value);
    }
    
    public int getScrollTopCol() {
        return this.scrollTopCol;
    }
    
    public void setScrollTopCol(final int value) {
        this.scrollTopCol = value;
    }
    
    public int getScrollTopRow() {
        return this.scrollTopRow;
    }
    
    public void setScrollTopRow(final int value) {
        this.scrollTopRow = value;
    }
    
    public int getScrollBottomCol() {
        return this.scrollBottomCol;
    }
    
    public void setScrollBottomCol(final int value) {
        this.scrollBottomCol = value;
    }
    
    public int getScrollBottomRow() {
        return this.scrollBottomRow;
    }
    
    public void setScrollBottomRow(final int value) {
        this.scrollBottomRow = value;
    }
    
    @Deprecated
    public int lockTopCol() {
        return this.scrollTopCol;
    }
    
    @Deprecated
    public void setLockTopCol(final int value) {
        this.scrollTopCol = value;
    }
    
    @Deprecated
    public int lockTopRow() {
        return this.scrollTopRow;
    }
    
    @Deprecated
    public void setLockTopRow(final int value) {
        this.scrollTopRow = value;
    }
    
    @Deprecated
    public int lockBottomCol() {
        return this.scrollBottomCol;
    }
    
    @Deprecated
    public void setLockBottomCol(final int value) {
        this.scrollBottomCol = value;
    }
    
    @Deprecated
    public int lockBottomRow() {
        return this.scrollBottomRow;
    }
    
    @Deprecated
    public void setLockBottomRow(final int value) {
        this.scrollBottomRow = value;
    }
    
    public int curCol() {
        return this.curCol;
    }
    
    public void setCurCol(final int value) {
        this.curCol = value;
    }
    
    public int curRow() {
        return this.curRow;
    }
    
    public void setCurRow(final int value) {
        this.curRow = value;
    }
    
    public GridFieldList merges() {
        return this.rects;
    }
    
    public IntList cellColors() {
        return this.cellColors;
    }
    
    public IntList fontColors() {
        return this.fontColors;
    }
    
    public IntList edgeColors() {
        return this.edgeColors;
    }
    
    public void setEdgeColor(final int index, final int color) {
        if (index >= 0 && index < 16) {
            this.edgeColors.set(index, color);
        }
    }
    
    public GridFontNames fontNames() {
        return this.fontNames;
    }
    
    public IntList fontSizes() {
        return this.fontSizes;
    }
    
    public boolean cellMerged(final int col, final int row) {
        return this.rects.getMergeRect(col, row) != null;
    }
    
    public CellField expandCell(final int col, final int row) {
        final CellField rt = this.rects.getMergeRect(col, row);
        return (CellField)((rt == null) ? new CellField(col, row, col, row) : rt.clone());
    }
    
    public boolean mergeCells(final int left, final int top, final int right, final int bottom) {
        final CellField cf = new CellField(left, top, right, bottom);
        return this.rects.addMergeRect(cf);
    }
    
    public LabelItems getLabels() {
        return this.labels;
    }
    
    private void clear() {
        this.scrollTopCol = -1;
        this.scrollTopRow = -1;
        this.scrollBottomCol = -1;
        this.scrollBottomRow = -1;
        this.curCol = -1;
        this.curRow = -1;
        this.labels.clear();
        this.extDatas.clear();
        this.cellCSS = null;
        this.cellScript = null;
    }
    
    public void loadFromStream(final Stream stream) throws StreamException {
        final ValueList valueList = new ValueList();
        final ValueList cellCSSList = new ValueList();
        final ValueList cellScriptList = new ValueList();
        final WordList propIndexs = new WordList();
        final KeyValues dataProps = new KeyValues();
        final BIFF b = new BIFF();
        b.data().setUseEncode(stream.getUseEncode());
        b.data().setCharset(stream.getCharset());
        final long p = stream.getPosition();
        BIFF.readBIFF(1, b, stream);
        String signFlag = b.data().readString(4);
        if (!"BIFF".equals(signFlag)) {
            stream.setPosition(p);
            BIFF.readBIFF(2, b, stream);
            signFlag = b.data().readString(4);
            if (!"BIFF".equals(signFlag)) {
                throw new StreamException("错误的文件格式");
            }
        }
        final int ver = b.data().readInt();
        int fileVer;
        if (b.data().getPosition() < b.data().getSize()) {
            fileVer = b.data().readInt();
        }
        else {
            fileVer = 16777216;
        }
        this.clear();
        this.useFontSize = false;
        while (b.ident != 127) {
            BIFF.readBIFF(ver, b, stream);
            switch (b.ident) {
                case 20: {
                    this.labels.loadFromStream(b.data());
                    if (fileVer == 16777216) {
                        this.labels.ver10Tover11();
                        continue;
                    }
                    continue;
                }
                default: {
                    continue;
                }
                case 25: {
                    dataProps.loadFromStream(b.data());
                    continue;
                }
                case 18: {
                    this.useFontSize = true;
                    continue;
                }
                case 6: {
                    this.fontNames.loadFromStream(b.data());
                    continue;
                }
                case 8: {
                    this.fontSizes.loadFromStream(b.data(), b.size);
                    continue;
                }
                case 7: {
                    this.fontColors.loadFromStream(b.data(), b.size);
                    continue;
                }
                case 10: {
                    this.cellColors.loadFromStream(b.data(), b.size);
                    continue;
                }
                case 9: {
                    this.edgeColors.loadFromStream(b.data(), b.size);
                    continue;
                }
                case 2: {
                    this.colWidths.loadFromStream(b.data(), b.size);
                    continue;
                }
                case 13: {
                    this.colAutoSize.loadFromStream(b.data(), b.size);
                    continue;
                }
                case 12: {
                    this.colVisible.loadFromStream(b.data(), b.size);
                    continue;
                }
                case 21: {
                    this.hideDataCol.loadFromStream(b.data(), b.size);
                    continue;
                }
                case 22: {
                    this.hideDataRow.loadFromStream(b.data(), b.size);
                    continue;
                }
                case 3: {
                    this.rowHeights.loadFromStream(b.data(), b.size);
                    continue;
                }
                case 15: {
                    this.rowAutoSize.loadFromStream(b.data(), b.size);
                    continue;
                }
                case 14: {
                    this.rowVisible.loadFromStream(b.data(), b.size);
                    continue;
                }
                case 4: {
                    valueList.loadFromStream(b.data());
                    continue;
                }
                case 11: {
                    this.propList.loadFromStreamEx(b.data(), 0, 12);
                    continue;
                }
                case 19: {
                    this.propList.loadFromStreamEx(b.data(), 12, 8);
                    continue;
                }
                case 1: {
                    propIndexs.loadFromStream(b.data(), b.size);
                    continue;
                }
                case 5: {
                    this.rects.loadFromStream(b.data(), b.size);
                    continue;
                }
                case 16: {
                    this.scrollTopCol = b.data().readInt();
                    this.scrollTopRow = b.data().readInt();
                    this.scrollBottomCol = b.data().readInt();
                    this.scrollBottomRow = b.data().readInt();
                    continue;
                }
                case 17: {
                    this.curCol = b.data().readInt();
                    this.curRow = b.data().readInt();
                    continue;
                }
                case 23: {
                    this.colCountPerPage = b.data().readInt();
                    this.rowCountPerPage = b.data().readInt();
                    this.rowPagingFirst = b.data().readBool();
                    continue;
                }
                case 24: {
                    cellCSSList.loadFromStream(b.data());
                    continue;
                }
                case 26: {
                    cellScriptList.loadFromStream(b.data());
                    continue;
                }
                case 27: {
                    this.extDatas.loadFromStream(b.data(), b.size);
                    continue;
                }
            }
        }
        this.gridDataClass = dataProps.getValue("gridclass");
        this.referenceBase = dataProps.getValue("gridrefbase");
        this.horzOffset = dataProps.getAsInt("horzoffset");
        this.vertOffset = dataProps.getAsInt("vertoffset");
        this.horzCenter = dataProps.getAsBoolean("horzcenter");
        this.vertCenter = dataProps.getAsBoolean("vertcenter");
        this.setColCount(this.colWidths.count());
        this.setRowCount(this.rowHeights.count());
        this.cells.resetSize(this.colWidths.count(), this.rowHeights.count());
        this.objects.resetSize(this.colWidths.count(), this.rowHeights.count());
        this.values.resetSize(this.colWidths.count(), this.rowHeights.count());
        int i = 0;
        for (int c = 0; c < this.colWidths.count(); ++c) {
            for (int r = 0; r < this.rowHeights.count(); ++r) {
                this.cells.setElement(c, r, new Integer(propIndexs.get(i)));
                ++i;
            }
        }
        for (i = 0; i < valueList.count(); ++i) {
            this.values.setElement(valueList.col(i), valueList.row(i), valueList.get(i));
        }
        for (i = 0; i < cellCSSList.count(); ++i) {
            this.setCellClass(cellCSSList.col(i), cellCSSList.row(i), cellCSSList.get(i));
        }
        for (i = 0; i < cellScriptList.count(); ++i) {
            this.setCellScript(cellScriptList.col(i), cellScriptList.row(i), cellScriptList.get(i));
        }
    }
    
    public void saveToStream(final Stream stream) throws StreamException {
        final ValueList valueList = new ValueList();
        final ValueList cellCSSList = new ValueList();
        final ValueList cellScriptList = new ValueList();
        final WordList propIndexs = new WordList();
        final KeyValues dataProps = new KeyValues();
        dataProps.setValue("gridclass", this.gridDataClass);
        dataProps.setValue("gridrefbase", this.referenceBase);
        dataProps.setAsInt("horzoffset", this.horzOffset);
        dataProps.setAsInt("vertoffset", this.vertOffset);
        dataProps.setAsBoolean("horzcenter", this.horzCenter);
        dataProps.setAsBoolean("vertcenter", this.vertCenter);
        for (int c = 0; c < this.colWidths.count(); ++c) {
            for (int r = 0; r < this.rowHeights.count(); ++r) {
                String cell = (String)this.values.getElement(c, r);
                if (cell != null && cell.length() > 0) {
                    final CellValue cv = new CellValue();
                    cv.col = c;
                    cv.row = r;
                    cv.value = cell;
                    valueList.add(cv);
                }
                Integer propIndex = (Integer)this.cells.getElement(c, r);
                if (propIndex == null) {
                    int pi = this.propList.indexOf((c == 0 || r == 0) ? GridConsts.DEF_HEADCELL_PROP : GridConsts.DEF_CELL_PROP);
                    if (c == 0 || r == 0) {
                        pi = this.propList.add(GridConsts.DEF_HEADCELL_PROP);
                    }
                    else {
                        pi = this.propList.add(GridConsts.DEF_CELL_PROP);
                    }
                    propIndex = new Integer(pi);
                }
                propIndexs.add(propIndex);
                cell = this.getCellClass(c, r);
                if (cell != null && cell.length() > 0) {
                    final CellValue cv2 = new CellValue();
                    cv2.col = c;
                    cv2.row = r;
                    cv2.value = cell;
                    cellCSSList.add(cv2);
                }
                cell = this.getCellScript(c, r);
                if (cell != null && cell.length() != 0) {
                    final CellValue cv2 = new CellValue();
                    cv2.col = c;
                    cv2.row = r;
                    cv2.value = cell;
                    cellScriptList.add(cv2);
                }
            }
        }
        final BIFF b = new BIFF();
        b.ident = 0;
        b.data().writeString("BIFF");
        b.data().writeInt(2);
        b.data().writeInt(16842752);
        BIFF.writeBIFF(1, b, stream);
        b.reset();
        b.ident = 25;
        dataProps.saveToStream(b.data());
        BIFF.writeBIFF(25, b, stream);
        b.reset();
        b.ident = 18;
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 6;
        this.fontNames.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 8;
        this.fontSizes.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 7;
        this.fontColors.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 10;
        this.cellColors.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 9;
        this.edgeColors.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 2;
        this.colWidths.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 13;
        this.colAutoSize.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 12;
        this.colVisible.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 21;
        this.hideDataCol.saveToStream(b.data());
        BIFF.writeBIFF(21, b, stream);
        b.reset();
        b.ident = 3;
        this.rowHeights.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 15;
        this.rowAutoSize.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 14;
        this.rowVisible.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 22;
        this.hideDataRow.saveToStream(b.data());
        BIFF.writeBIFF(22, b, stream);
        b.reset();
        b.ident = 4;
        valueList.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 11;
        this.propList.saveToStreamEx(b.data(), 0, 12);
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 19;
        this.propList.saveToStreamEx(b.data(), 12, 8);
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 1;
        propIndexs.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 5;
        this.rects.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 16;
        b.data().writeInt(this.scrollTopCol);
        b.data().writeInt(this.scrollTopRow);
        b.data().writeInt(this.scrollBottomCol);
        b.data().writeInt(this.scrollBottomRow);
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 17;
        b.data().writeInt(this.curCol);
        b.data().writeInt(this.curRow);
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 20;
        this.labels.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 23;
        b.data().writeInt(this.colCountPerPage);
        b.data().writeInt(this.rowCountPerPage);
        b.data().writeBool(this.rowPagingFirst);
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 24;
        cellCSSList.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 26;
        cellScriptList.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 27;
        this.extDatas.saveToStream(b.data());
        BIFF.writeBIFF(2, b, stream);
        b.reset();
        b.ident = 127;
        BIFF.writeBIFF(2, b, stream);
    }
    
    public void loadFromStream(final InputStream inStream) throws StreamException {
        final MemStream stream = new MemStream();
        try {
            stream.loadFromStream(inStream);
        }
        catch (final IOException e) {
            throw new StreamException((Throwable)e);
        }
        stream.setPosition(0L);
        this.loadFromStream((Stream)stream);
    }
    
    public void saveToStream(final OutputStream outStream) throws StreamException {
        final MemStream stream = new MemStream();
        this.saveToStream((Stream)stream);
        try {
            stream.saveToStream(outStream);
        }
        catch (final IOException e) {
            throw new StreamException((Throwable)e);
        }
    }
    
    public void loadFromFile(final String fileName) throws StreamException {
        final MemStream stream = new MemStream();
        try {
            stream.loadFromFile(fileName);
        }
        catch (final IOException e) {
            throw new StreamException((Throwable)e);
        }
        stream.setPosition(0L);
        this.loadFromStream((Stream)stream);
    }
    
    public void saveToFile(final String fileName) throws StreamException {
        final MemStream stream = new MemStream();
        this.saveToStream((Stream)stream);
        try {
            stream.saveToFile(fileName);
        }
        catch (final IOException e) {
            throw new StreamException((Throwable)e);
        }
    }
    
    public GridData makeCopy() throws StreamException {
        final Stream s = (Stream)new MemStream();
        this.saveToStream(s);
        s.setPosition(0L);
        final GridData gd = new GridData();
        gd.loadFromStream(s);
        gd.setOptions(this.getOptions());
        return gd;
    }
    
    public Object clone() {
        try {
            return this.makeCopy();
        }
        catch (final StreamException ex) {
            return null;
        }
    }
    
    public boolean copyFrom(final GridData srcData, final int fromLeft, final int fromTop, final int fromRight, final int fromBottom, final int toLeft, final int toTop) {
        final CellField dest = new CellField(toLeft, toTop, toLeft + fromRight - fromLeft, toTop + fromBottom - fromTop);
        if (!this.checkCol(dest.left, false)) {
            return false;
        }
        if (!this.checkCol(dest.right, false)) {
            return false;
        }
        if (!this.checkRow(dest.top, false)) {
            return false;
        }
        if (!this.checkRow(dest.bottom, false)) {
            return false;
        }
        if (!srcData.checkCol(fromLeft, false)) {
            return false;
        }
        if (!srcData.checkCol(fromRight, false)) {
            return false;
        }
        if (!srcData.checkRow(fromTop, false)) {
            return false;
        }
        if (!srcData.checkRow(fromBottom, false)) {
            return false;
        }
        this.rects.removeMergeRect(dest);
        final GridCopyConvertor convertor = (this == srcData) ? null : new GridCopyConvertor(srcData);
        for (int x = fromLeft; x <= fromRight; ++x) {
            for (int y = fromTop; y <= fromBottom; ++y) {
                final CellField field = srcData.expandCell(x, y);
                final String text = srcData.getCellData(field.left, field.top);
                final GridCell cell = srcData.getCellEx(field.left, field.top);
                if (field.left < fromLeft) {
                    field.left = fromLeft;
                }
                if (field.right > fromRight) {
                    field.right = fromRight;
                }
                if (field.top < fromTop) {
                    field.top = fromTop;
                }
                if (field.bottom > fromBottom) {
                    field.bottom = fromBottom;
                }
                if (field.left >= x) {
                    if (field.top >= y) {
                        final int xx = toLeft + x - fromLeft;
                        final int yy = toTop + y - fromTop;
                        if (convertor != null) {
                            convertor.convert(cell);
                        }
                        this.setCell(xx, yy, cell);
                        this.setCellData(xx, yy, text);
                        this.setObj(xx, yy, srcData.getObj(field.left, field.top));
                        this.setCellClass(xx, yy, srcData.getCellClass(field.left, field.top));
                        this.setCellScript(xx, yy, srcData.getCellScript(field.left, field.top));
                        this.mergeCells(xx, yy, toLeft + field.right - fromLeft, toTop + field.bottom - fromTop);
                    }
                }
            }
        }
        if (dest.top == 1 && dest.bottom == this.getRowCount() - 1) {
            for (int x = fromLeft; x <= fromRight; ++x) {
                final int xx2 = toLeft + x - fromLeft;
                this.setHideDataCol(xx2, srcData.getHideDataCol(x));
                this.setColAutoSize(xx2, srcData.getColAutoSize(x));
                this.setColWidths(xx2, srcData.getColWidths(x));
                this.setColVisible(xx2, srcData.getColVisible(x));
            }
        }
        if (dest.left == 1 && dest.right == this.getColCount() - 1) {
            for (int y2 = fromTop; y2 <= fromBottom; ++y2) {
                final int yy2 = toTop + y2 - fromTop;
                this.setHideDataRow(yy2, srcData.getHideDataRow(y2));
                this.setRowAutoSize(yy2, srcData.getRowAutoSize(y2));
                this.setRowHeights(yy2, srcData.getRowHeights(y2));
                this.setRowVisible(yy2, srcData.getRowVisible(y2));
            }
        }
        return true;
    }
    
    public int getColCountPerPage() {
        return this.colCountPerPage;
    }
    
    public void setColCountPerPage(final int value) {
        this.colCountPerPage = value;
    }
    
    public int getRowCountPerPage() {
        return this.rowCountPerPage;
    }
    
    public void setRowCountPerPage(final int value) {
        this.rowCountPerPage = value;
    }
    
    public boolean getRowPagingFirst() {
        return this.rowPagingFirst;
    }
    
    public void setRowPagingFirst(final boolean value) {
        this.rowPagingFirst = value;
    }
    
    public int getColPageCount() {
        if (this.colCountPerPage <= 0) {
            return 1;
        }
        int colCount = this.getColCount() - 1;
        if (this.scrollBottomCol > 0 && this.scrollBottomCol > this.scrollTopCol) {
            colCount = this.scrollBottomCol;
        }
        if (this.scrollTopCol > 0) {
            colCount -= this.scrollTopCol - 1;
        }
        return (colCount - 1) / this.colCountPerPage + 1;
    }
    
    public int getRowPageCount() {
        if (this.rowCountPerPage <= 0) {
            return 1;
        }
        int rowCount = this.getRowCount() - 1;
        if (this.scrollBottomRow > 0 && this.scrollBottomRow > this.scrollTopRow) {
            rowCount = this.scrollBottomRow;
        }
        if (this.scrollTopRow > 0) {
            rowCount -= this.scrollTopRow - 1;
        }
        return (rowCount - 1) / this.rowCountPerPage + 1;
    }
    
    public int getPageCount() {
        return this.getColPageCount() * this.getRowPageCount();
    }
    
    public GridData getPagedGridData(final int pageIndex) {
        int px = 0;
        int py = 0;
        if (this.rowPagingFirst) {
            final int count = this.getRowPageCount();
            px = pageIndex / count;
            py = pageIndex % count;
        }
        else {
            final int count = this.getColPageCount();
            px = pageIndex % count;
            py = pageIndex / count;
        }
        return this.getPagedGridData(px, py);
    }
    
    public GridData getPagedGridData(final int colPageIndex, final int rowPageIndex) {
        if (colPageIndex < 0 || rowPageIndex < 0) {
            return null;
        }
        int colHead = this.scrollTopCol - 1;
        int colTail = this.scrollBottomCol + 1;
        int rowHead = this.scrollTopRow - 1;
        int rowTail = this.scrollBottomRow + 1;
        if (colHead < 0) {
            colHead = 0;
        }
        if (colTail <= colHead) {
            colTail = this.getColCount();
        }
        if (rowHead < 0) {
            rowHead = 0;
        }
        if (rowTail <= rowHead) {
            rowTail = this.getRowCount();
        }
        int colCountPerPage = this.colCountPerPage;
        if (colCountPerPage <= 0) {
            colCountPerPage = colTail - colHead - 1;
        }
        int rowCountPerPage = this.rowCountPerPage;
        if (rowCountPerPage <= 0) {
            rowCountPerPage = rowTail - rowHead - 1;
        }
        final int dataColCount = Math.min(colCountPerPage, colTail - colPageIndex * colCountPerPage - colHead - 1);
        if (dataColCount <= 0 && colPageIndex != 0) {
            return null;
        }
        final int fromDataCol = colHead + 1 + colPageIndex * colCountPerPage;
        final int dataRowCount = Math.min(rowCountPerPage, rowTail - rowPageIndex * rowCountPerPage - rowHead - 1);
        if (dataRowCount <= 0 && rowPageIndex != 0) {
            return null;
        }
        final int fromDataRow = rowHead + 1 + rowPageIndex * rowCountPerPage;
        if (1 + colHead + dataColCount == colTail && 1 + rowHead + dataRowCount == rowTail) {
            return this;
        }
        final GridData result = new GridData(this.getOptions());
        result.cellColors = (IntList)this.cellColors.clone();
        result.edgeColors = (IntList)this.edgeColors.clone();
        result.fontNames = (GridFontNames)this.fontNames.clone();
        result.fontColors = (IntList)this.fontColors.clone();
        result.fontSizes = (IntList)this.fontSizes.clone();
        result.setDataClass(this.getDataClass());
        result.setColCount(1 + colHead + dataColCount + this.getColCount() - colTail);
        result.setRowCount(1 + rowHead + dataRowCount + this.getRowCount() - rowTail);
        result.setLockTopCol(colHead + 1);
        result.setLockTopRow(rowHead + 1);
        result.setLockBottomCol(colHead + dataColCount);
        result.setLockBottomRow(rowHead + dataRowCount);
        for (int i = 1; i <= colHead; ++i) {
            result.setColWidths(i, this.getColWidths(i));
            result.setColAutoSize(i, this.getColAutoSize(i));
            result.setColVisible(i, this.getColVisible(i));
            result.setHideDataCol(i, this.getHideDataCol(i));
        }
        for (int i = 0; i < dataColCount; ++i) {
            result.setColWidths(colHead + 1 + i, this.getColWidths(fromDataCol + i));
            result.setColAutoSize(colHead + 1 + i, this.getColAutoSize(fromDataCol + i));
            result.setColVisible(colHead + 1 + i, this.getColVisible(fromDataCol + i));
            result.setHideDataCol(colHead + 1 + i, this.getHideDataCol(fromDataCol + i));
        }
        for (int i = colTail; i < this.getColCount(); ++i) {
            result.setColWidths(colHead + 1 + dataColCount + i - colTail, this.getColWidths(i));
            result.setColAutoSize(colHead + 1 + dataColCount + i - colTail, this.getColAutoSize(i));
            result.setColVisible(colHead + 1 + dataColCount + i - colTail, this.getColVisible(i));
            result.setHideDataCol(colHead + 1 + dataColCount + i - colTail, this.getHideDataCol(i));
        }
        for (int i = 1; i <= rowHead; ++i) {
            result.setRowHeights(i, this.getRowHeights(i));
            result.setRowAutoSize(i, this.getRowAutoSize(i));
            result.setRowVisible(i, this.getRowVisible(i));
            result.setHideDataRow(i, this.getHideDataRow(i));
        }
        for (int i = 0; i < dataRowCount; ++i) {
            result.setRowHeights(rowHead + 1 + i, this.getRowHeights(fromDataRow + i));
            result.setRowAutoSize(rowHead + 1 + i, this.getRowAutoSize(fromDataRow + i));
            result.setRowVisible(rowHead + 1 + i, this.getRowVisible(fromDataRow + i));
            result.setHideDataRow(rowHead + 1 + i, this.getHideDataRow(fromDataRow + i));
        }
        for (int i = rowTail; i < this.getRowCount(); ++i) {
            result.setRowHeights(rowHead + 1 + dataRowCount + i - rowTail, this.getRowHeights(i));
            result.setRowAutoSize(rowHead + 1 + dataRowCount + i - rowTail, this.getRowAutoSize(i));
            result.setRowVisible(rowHead + 1 + dataRowCount + i - rowTail, this.getRowVisible(i));
            result.setHideDataRow(rowHead + 1 + dataRowCount + i - rowTail, this.getHideDataRow(i));
        }
        if (colHead > 0 && rowHead > 0) {
            result.copyFrom(this, 1, 1, colHead, rowHead, 1, 1);
        }
        if (dataColCount > 0 && rowHead > 0) {
            result.copyFrom(this, fromDataCol, 1, fromDataCol + dataColCount - 1, rowHead, colHead + 1, 1);
        }
        if (colTail < this.getColCount() && rowHead > 0) {
            result.copyFrom(this, colTail, 1, this.getColCount() - 1, rowHead, colHead + 1 + dataColCount, 1);
        }
        if (colHead > 0 && dataRowCount > 0) {
            result.copyFrom(this, 1, fromDataRow, colHead, fromDataRow + dataRowCount - 1, 1, rowHead + 1);
        }
        if (dataColCount > 0 && dataRowCount > 0) {
            result.copyFrom(this, fromDataCol, fromDataRow, fromDataCol + dataColCount - 1, fromDataRow + dataRowCount - 1, colHead + 1, rowHead + 1);
        }
        if (colTail < this.getColCount() && dataRowCount > 0) {
            result.copyFrom(this, colTail, fromDataRow, this.getColCount() - 1, fromDataRow + dataRowCount - 1, colHead + 1 + dataColCount, rowHead + 1);
        }
        if (colHead > 0 && rowTail < this.getRowCount()) {
            result.copyFrom(this, 1, rowTail, colHead, this.getRowCount() - 1, 1, rowHead + 1 + dataRowCount);
        }
        if (dataColCount > 0 && rowTail < this.getRowCount()) {
            result.copyFrom(this, fromDataCol, rowTail, fromDataCol + dataColCount - 1, this.getRowCount() - 1, colHead + 1, rowHead + 1 + dataRowCount);
        }
        if (colTail < this.getColCount() && rowTail < this.getRowCount()) {
            result.copyFrom(this, colTail, rowTail, this.getColCount() - 1, this.getRowCount() - 1, colHead + 1 + dataColCount, rowHead + 1 + dataRowCount);
        }
        return result;
    }
    
    public int getDataRowCount() {
        final int top = (this.scrollTopRow > 0) ? this.scrollTopRow : 1;
        final int bottom = (this.scrollBottomRow > 0) ? this.scrollBottomRow : (this.getRowCount() - 1);
        int count = 0;
        for (int i = top; i <= bottom; ++i) {
            if (this.getRowVisible(i)) {
                ++count;
            }
        }
        return count;
    }
    
    public int getDataColCount() {
        final int left = (this.scrollTopCol > 0) ? this.scrollTopCol : 1;
        final int right = (this.scrollBottomCol > 0) ? this.scrollBottomCol : (this.getColCount() - 1);
        int count = 0;
        for (int i = left; i <= right; ++i) {
            if (this.getColVisible(i)) {
                ++count;
            }
        }
        return count;
    }
    
    public ExtendedDatas getExtDatas() {
        return this.extDatas;
    }
    
    public int getHorzOffset() {
        return this.horzOffset;
    }
    
    public void setHorzOffset(final int offset) {
        this.horzOffset = offset;
    }
    
    public int getVertOffset() {
        return this.vertOffset;
    }
    
    public void setVertOffset(final int offset) {
        this.vertOffset = offset;
    }
    
    public boolean getHorzCenter() {
        return this.horzCenter;
    }
    
    public void setHorzCenter(final boolean value) {
        this.horzCenter = value;
    }
    
    public boolean getVertCenter() {
        return this.vertCenter;
    }
    
    public void setVertCenter(final boolean value) {
        this.vertCenter = value;
    }
    
    public static GridData bytesToGrid(final byte[] data) {
        if (data == null) {
            return null;
        }
        GridData gd = new GridData();
        try {
            final Stream s = (Stream)new MemStream();
            s.writeBuffer(data, 0, data.length);
            s.setPosition(0L);
            gd.loadFromStream(s);
        }
        catch (final Exception ex) {
            ex.printStackTrace();
            gd = null;
        }
        return gd;
    }
    
    public static GridData bytesToGrid(final byte[] data, final String encodeCharset) {
        if (data == null) {
            return null;
        }
        GridData gd = new GridData();
        try {
            final Stream s = (Stream)new MemStream();
            s.setUseEncode(true);
            s.setCharset(encodeCharset);
            s.writeBuffer(data, 0, data.length);
            s.setPosition(0L);
            gd.loadFromStream(s);
        }
        catch (final Exception ex) {
            ex.printStackTrace();
            gd = null;
        }
        return gd;
    }
    
    public static byte[] gridToBytes(final GridData grid) {
        if (grid == null) {
            return null;
        }
        final Stream store = (Stream)new MemStream();
        try {
            grid.saveToStream(store);
            return store.getBytes();
        }
        catch (final StreamException ex) {
            ex.printStackTrace();
            return null;
        }
    }
    
    public static byte[] gridToBytes(final GridData grid, final String encodeCharset) {
        if (grid == null) {
            return null;
        }
        final Stream store = (Stream)new MemStream();
        try {
            store.setUseEncode(true);
            store.setCharset(encodeCharset);
            grid.saveToStream(store);
            return store.getBytes();
        }
        catch (final StreamException ex) {
            ex.printStackTrace();
            return null;
        }
    }
    
    public static GridData base64ToGrid(final String base64Str) {
        final byte[] data = Base64.base64ToByteArray(base64Str);
        return bytesToGrid(data);
    }
    
    public static GridData base64ToGrid(final String base64Str, final String encodeCharset) {
        final byte[] data = Base64.base64ToByteArray(base64Str);
        return bytesToGrid(data, encodeCharset);
    }
    
    public static String gridToBase64(final GridData grid) {
        final byte[] data = gridToBytes(grid);
        return Base64.byteArrayToBase64(data);
    }
    
    public static String gridToBase64(final GridData grid, final String encodeCharset) {
        final byte[] data = gridToBytes(grid, encodeCharset);
        return Base64.byteArrayToBase64(data);
    }
    
    public void convertData(final CellConvertor convertor) {
        for (int col = 0; col < this.getColCount(); ++col) {
            for (int row = 0; row < this.getRowCount(); ++row) {
                final CellField cf = this.expandCell(col, row);
                if (cf.left == col && cf.top == row) {
                    final GridCell cell = this.getCellEx(col, row);
                    if (convertor.convert(this, cell)) {
                        this.setCell(col, row, cell);
                    }
                }
            }
        }
    }
    
    public int getOptions() {
        return this.options;
    }
    
    public void setOptions(final int value) {
        this.options = value;
        this.rects.setAutoMerge((value & 0x1) != 0x0);
    }
    
    private final class GridCopyConvertor
    {
        GridData src;
        int[] fontNameMaps;
        int[] fontSizeMaps;
        int[] fontColorMaps;
        int[] cellColorMaps;
        int[] edgeColorMaps;
        private Stack freeFontNames;
        private Stack freeFontSizes;
        private Stack freeFontColors;
        private Stack freeCellColors;
        private Stack freeEdgeColors;
        
        public GridCopyConvertor(final GridData src) {
            this.src = src;
            this.mergeFontNames(src.fontNames);
            this.fontSizeMaps = this.mergeList(src.fontSizes, GridData.this.fontSizes);
            this.fontColorMaps = this.mergeList(src.fontColors, GridData.this.fontColors);
            this.cellColorMaps = this.mergeList(src.cellColors, GridData.this.cellColors);
            this.edgeColorMaps = this.mergeList(src.edgeColors, GridData.this.edgeColors);
        }
        
        private void mergeFontNames(final GridFontNames srcFontNames) {
            boolean diff = false;
            final int[] maps = new int[srcFontNames.count()];
            for (int i = 0; i < srcFontNames.count(); ++i) {
                final int index = GridData.this.fontNames.indexOfFont(srcFontNames.getFont(i));
                if (i != (maps[i] = index)) {
                    diff = true;
                }
            }
            this.fontNameMaps = (int[])(diff ? maps : null);
        }
        
        private int[] mergeList(final IntList srcList, final IntList destList) {
            boolean diff = false;
            final int[] maps = new int[srcList.count()];
            for (int i = 0; i < srcList.count(); ++i) {
                final int index = destList.indexOf(srcList.get(i));
                if (i != (maps[i] = index)) {
                    diff = true;
                }
            }
            return (int[])(diff ? maps : null);
        }
        
        public void convert(final GridCell cell) {
            final byte[] data = cell.getPropData();
            if (this.fontNameMaps != null) {
                final int index = data[4] & 0x3F;
                final int newIndex = this.mapFontName(index);
                data[4] = (byte)((data[4] & 0xC0) | (newIndex & 0x3F));
            }
            if (this.fontSizeMaps != null) {
                final int index = data[5] & 0x3F;
                final int newIndex = this.mapFontSize(index);
                data[5] = (byte)((data[5] & 0xC0) | (newIndex & 0x3F));
            }
            if (this.fontColorMaps != null) {
                final int index = data[6] >> 0 & 0x3F;
                final int newIndex = this.mapFontColor(index);
                data[6] = (byte)((data[6] & 0xC0) | (newIndex & 0x3F));
            }
            if (this.cellColorMaps != null) {
                final int index = data[0];
                final int newIndex = this.mapCellColor(index);
                data[0] = (byte)newIndex;
            }
            if (this.edgeColorMaps != null) {
                int index = data[9] >> 4 & 0xF;
                int newIndex = this.mapEdgeColor(index);
                data[9] = (byte)((data[9] & 0xF) | (newIndex << 4 & 0xF0));
                index = (data[10] >> 4 & 0xF);
                newIndex = this.mapEdgeColor(index);
                data[10] = (byte)((data[10] & 0xF) | (newIndex << 4 & 0xF0));
            }
        }
        
        private int mapCellColor(final int index) {
            int newIndex;
            if (this.cellColorMaps[index] == -1) {
                if (this.freeCellColors().isEmpty()) {
                    newIndex = ((GridData.this.cellColors.count() >= 64) ? 0 : GridData.this.cellColors.add(this.src.cellColors.get(index)));
                }
                else {
                    newIndex = this.freeCellColors().pop();
                    GridData.this.cellColors.set(newIndex, this.src.cellColors.get(index));
                }
                this.cellColorMaps[index] = newIndex;
            }
            else {
                newIndex = this.cellColorMaps[index];
            }
            return newIndex;
        }
        
        private int mapFontColor(final int index) {
            int newIndex;
            if (this.fontColorMaps[index] == -1) {
                if (this.freeFontColors().isEmpty()) {
                    newIndex = ((GridData.this.fontColors.count() >= 64) ? 0 : GridData.this.fontColors.add(this.src.fontColors.get(index)));
                }
                else {
                    newIndex = this.freeFontColors().pop();
                    GridData.this.fontColors.set(newIndex, this.src.fontColors.get(index));
                }
                this.fontColorMaps[index] = newIndex;
            }
            else {
                newIndex = this.fontColorMaps[index];
            }
            return newIndex;
        }
        
        private int mapFontSize(final int index) {
            int newIndex;
            if (this.fontSizeMaps[index] == -1) {
                if (this.freeFontSizes().isEmpty()) {
                    newIndex = ((GridData.this.fontSizes.count() >= 64) ? 0 : GridData.this.fontSizes.add(this.src.fontSizes.get(index)));
                }
                else {
                    newIndex = this.freeFontSizes().pop();
                    GridData.this.fontSizes.set(newIndex, this.src.fontSizes.get(index));
                }
                this.fontSizeMaps[index] = newIndex;
            }
            else {
                newIndex = this.fontSizeMaps[index];
            }
            return newIndex;
        }
        
        private int mapFontName(final int index) {
            int newIndex;
            if (this.fontNameMaps[index] == -1) {
                if (this.freeFontNames().isEmpty()) {
                    newIndex = ((GridData.this.fontNames.count() >= 64) ? 0 : GridData.this.fontNames.addFont(this.src.fontNames.getFont(index)));
                }
                else {
                    newIndex = this.freeFontNames().pop();
                    GridData.this.fontNames.setFont(newIndex, this.src.fontNames.getFont(index));
                }
                this.fontNameMaps[index] = newIndex;
            }
            else {
                newIndex = this.fontNameMaps[index];
            }
            return newIndex;
        }
        
        private int mapEdgeColor(final int index) {
            int newIndex;
            if (this.edgeColorMaps[index] == -1) {
                if (this.freeEdgeColors().isEmpty()) {
                    newIndex = ((GridData.this.edgeColors.count() >= 64) ? 0 : GridData.this.edgeColors.add(this.src.edgeColors.get(index)));
                }
                else {
                    newIndex = this.freeEdgeColors().pop();
                    GridData.this.edgeColors.set(newIndex, this.src.edgeColors.get(index));
                }
                this.edgeColorMaps[index] = newIndex;
            }
            else {
                newIndex = this.edgeColorMaps[index];
            }
            return newIndex;
        }
        
        private Stack freeFontNames() {
            if (this.freeFontNames == null) {
                final boolean[] flags = new boolean[GridData.this.fontNames.count()];
                for (int i = 0; i < GridData.this.propList.count(); ++i) {
                    final byte[] data = GridData.this.propList.get(i);
                    final int index = data[4] & 0x3F;
                    flags[index] = true;
                }
                this.freeFontNames = new Stack();
                for (int i = 0; i < flags.length; ++i) {
                    if (!flags[i]) {
                        this.freeFontNames.push(new Integer(i));
                    }
                }
            }
            return this.freeFontNames;
        }
        
        private Stack freeFontSizes() {
            if (this.freeFontSizes == null) {
                final boolean[] flags = new boolean[GridData.this.fontSizes.count()];
                for (int i = 0; i < GridData.this.propList.count(); ++i) {
                    final byte[] data = GridData.this.propList.get(i);
                    final int index = data[5] & 0x3F;
                    flags[index] = true;
                }
                this.freeFontSizes = new Stack();
                for (int i = 0; i < flags.length; ++i) {
                    if (!flags[i]) {
                        this.freeFontSizes.push(new Integer(i));
                    }
                }
            }
            return this.freeFontSizes;
        }
        
        private Stack freeFontColors() {
            if (this.freeFontColors == null) {
                final boolean[] flags = new boolean[GridData.this.fontColors.count()];
                for (int i = 0; i < GridData.this.propList.count(); ++i) {
                    final byte[] data = GridData.this.propList.get(i);
                    final int index = data[6] >> 0 & 0x3F;
                    flags[index] = true;
                }
                this.freeFontColors = new Stack();
                for (int i = 0; i < flags.length; ++i) {
                    if (!flags[i]) {
                        this.freeFontColors.push(new Integer(i));
                    }
                }
            }
            return this.freeFontColors;
        }
        
        private Stack freeCellColors() {
            if (this.freeCellColors == null) {
                final boolean[] flags = new boolean[GridData.this.cellColors.count()];
                for (int i = 0; i < GridData.this.propList.count(); ++i) {
                    final byte[] data = GridData.this.propList.get(i);
                    flags[data[0]] = true;
                }
                this.freeCellColors = new Stack();
                for (int i = 0; i < flags.length; ++i) {
                    if (!flags[i]) {
                        this.freeCellColors.push(new Integer(i));
                    }
                }
            }
            return this.freeCellColors;
        }
        
        private Stack freeEdgeColors() {
            if (this.freeEdgeColors == null) {
                final boolean[] flags = new boolean[GridData.this.edgeColors.count()];
                for (int i = 0; i < GridData.this.propList.count(); ++i) {
                    final byte[] data = GridData.this.propList.get(i);
                    int index = data[9] >> 4 & 0xF;
                    flags[index] = true;
                    index = (data[10] >> 4 & 0xF);
                    flags[index] = true;
                }
                this.freeEdgeColors = new Stack();
                for (int i = 0; i < flags.length; ++i) {
                    if (!flags[i]) {
                        this.freeEdgeColors.push(new Integer(i));
                    }
                }
            }
            return this.freeEdgeColors;
        }
    }
}
